/**
 * This program is a library for solving science calculations.          *
 * Copyright 2008  Christopher Olah  christopherolah.co@gmail.com       *

 * This program is free software; you can redistribute it and/or        *
 * modify it under the terms of the GNU General Public License as       *
 * published by the Free Software Foundation; either version 2 of       *
 * the License, or (at your option) any later version.                  *

 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *

 * You should have received a copy of the GNU General Public License    *
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.*/

/**
* I use whitespace very liberaly. I suggest you edit and view this code on
* a big screen.
*/

/**
* @author Christopher Olah
* @todo fix ! (factorial) and @ (base), control structures
* Ideas:
* -Calculus
* -variable naming
*/


#include <iostream>
#include <cstdlib>
#include <string>
#include <math.h>
#include <map>

#include "science_math.h"
//#include "fstring.h"

using namespace std;

class solver {

	protected:

	/** @code remove all whitespace unless it is followed by a '\' in which case remove the '\'*/
	fstring clean_whitespace(fstring input){
		fstring pass;
		pass = input.get_till(' ');
		while (!input.atend()){
			input.skip();
			pass += input.get_till(' ');
			if (pass[pass.length()-1] == '\\') {
				pass = pass.get_till(pass.length() - 1);
				pass += ' ';
			} 
		}
		return control(pass);
	}

	/** @code control structures*/
	fstring control(fstring input) {
		if (!input.contains('{')) {
			return semicolon(input);
		}
		fstring out;
		while (!input.atend()){
			fstring structure = input.get_till('(');
			input.skip(); // pass ')'
			fstring argument = input.get_till(')');
			input.skip(); input.skip(); //pass ')' and '{'
			fstring data = input.get_till('}');
			input.skip();
			if      (structure == "if" || structure == "?"){
				if (bracket(argument) != 0){
					out+=control(data);
				}
			}else if(structure == "repeat"                ){
				for (int i = 0; i < bracket(argument); i++){
					out+=control(data);
				}
			}else if(structure == "while"                 ){
				while(bracket(argument) != 0){
					out+=control(data);
				}
			}else if(structure == "for"                   ){
				out+=control(argument.get_till(';'));
				argument.skip();
				fstring condition = argument.get_till(';');
				argument.skip();
				fstring changer = argument.get_rest();
				while(arithmatic(condition)){
					out+=control(data);
					out+=control(changer);
				}
			}else                                          {
				input.position = 0;
				out += semicolon(input.get_till(';') + ';');
				input.skip();
			}
			if (input.atend()){break;}
			input=input.get_rest();
		}
	return out;
	}


/** @code allow for multiple expressions on a line. Also, anything with out a semicolong is prepared with out= for conveninece. */
	fstring semicolon(fstring input){
		fstring out;
		fstring temporary;
		while (true){
			temporary = "";
			if (!input.contains(';')){temporary = "out=";} //we want to let simple things work without someone having to explicitly call them
			temporary+=input.get_till(';');
			out += assign(temporary);
			if (!input.atend()){input.skip();} //pass ';'if there is one
			if (input.atend()) {break;}        //avoid infinite loop...
			input=input.get_rest();            //shrink input...
		}
		return out;
	}

	fstring assign(fstring input){
		fstring out;
		fstring variable = input.get_till('='); //remember, theres always a variable, and it's always first, even if the user doesn't know... :)
		input.skip();//pass '='
		fstring value =  input.get_rest();
		if (value[0] == '"'){//Input between "s is protected...
			value.skip();
			value=value.get_rest(); //get rid of first "
			value=value.get_till(value.length()-2);//get rid of last "
		}else{
			value=plus_minus(value);
		}
		//What variable
		if      (variable    == "out" ){ out+=value + "\n";    } //out is a pseudo variable
		else if (variable    == "@"   ){ this->s_base=value;   } //s_base is a psedo variable
		else if (variable    == "var" ){ this->var=value.dbl();}
		else if (variable    == "F(x)"){ this->F=value;        }
		else if (variable[0] == 'v'   ){ 
			this->v[variable.get_from(1).itg()]=value.dbl();
		}
		return out;
	}

/** @code plus_minus returns two outputs*/

	fstring plus_minus(fstring input){
		fstring out;
		if (!input.contains('~')) {
			out = arithmatic(input);
		}else {
			fstring part1 = input.get_till('~');
			input.skip();
			fstring part2 = input.get_rest();
			out += plus_minus(part1 + "+" + part2).str() + "\n";
			out += plus_minus(part1 + "-" + part2).str();
		}
		return out;
	}

/** @code below are the arithmatic functions. They are reponsible for simplifying and solving equations. */

	double arithmatic(fstring input){return bracket(input);} //Pointer to first arimthmatic function. This allows fast insertion of methods.

/** @code Input is searched for brackets of which the content is retreived. The content is evaluated before anythin else.*/

	double bracket(fstring input) {
		fstring pass;
		while(!input.atend()){
			pass += input.get_till('(');
			if (!input.atend()) {
				pass += bracket(input.get_in('(',')'));
				input.skip();
			}
		}
		return minus(pass);
	}


/** @code minus does preprocessing for the minus sign. It prevents the need for subtraction by reimplimenting it as addition.
This model prevents scenarios like 1/-1 being translated as 1/0-1*/

	double minus(fstring input) {
		fstring pass = input.get_till('-');
		if (input.position == 0) {
			input.skip();
			pass += '-';
			pass += input.get_till('-');
		}
		while (!input.atend()) {
			char last = pass[pass.length()-1];
			if      (  last == '+'
			        || last == '*'
				    || last == '/'
			        || last == '^')  { input.skip(); pass += '-';                             }
			else if (  last == '-')  { input.skip(); pass  = pass.get_first(pass.length()-1); }
			else                     { input.skip(); pass += "+-";                            }
			pass += input.get_till('-');
		}
		return comperative(pass);
	}

/** @code Input is searched for == or != which are then applied... */

	double comperative(fstring input) {
		fstring part1, part2;
		part1 = input.get_till("==");
		if (input.atend()) { //if we found an ==, we won't be at the end
			input.position = 0;
			part1 = input.get_till("!=");
			if (input.atend()) {//if we found a !=, we won't be at the end
				input.position = 0;
				return addition(input);	
			}else {
				input.skip(); input.skip(); //pass !=
				part2 = input.get_rest();
				return (addition(part1) != addition(part2));
			}
		}else {
			input.skip(); input.skip(); //pass ==
			part2 = input.get_rest();
			return (addition(part1) == addition(part2));
		}
	}

/**
@code
Input is parsed by having individual functions for each symbol.
Each one gets the content between instances of these sybols,
and passes it the next function. (addition->multiplication->division->exponent)
It applies the symbols corresponding mathamatical function to the return.
It returns the result as a double.
Exponent passes to special where the string is actually parsed to a double.
*/

	double addition(fstring input) {
		double value = multiplication(input.get_till('+'));
		while (!input.atend()) {
			input.skip(); //go past '+'
			value += multiplication(input.get_till('+'));
		}
		return value;
	}

	double multiplication(fstring input) {
		double value;
		value = division(input.get_till('*'));
		while (!input.atend()) {
			input.skip(); //go past '*'
			value *= division(input.get_till('*'));
		}
		return value;
	}

	double division(fstring input) {
		double value;
		value = exponent(input.get_till('/'));
		while (!input.atend()) {
			input.skip(); //go past '/'
			value /= exponent(input.get_till('/'));
		}
		return value;
	}

/**
@code: Because there can ge only one exponent sign, the exponent function is different from addition, subtraction, multiplication and division methods. Additionaly, we must use an if to prevent agains base^special("").
*/
	double exponent(fstring input) {
		double base, power;
		base = special(input.get_till('^'));
		if (input.atend()){
			return base;
		}else {
			input.skip(); //go past '^'
			return pow(base, special(input.get_till('^')));
		}
	}

/**
This is where you should add special variables and functions

@code: special is what transforms fstrings into actual numbers. In addition to providing these numbers, it also provides many variables and functions. It is recursive: a function will parse its input with special. This means that functions and variable chains are executed left to write. If a string doesn't match anythin, it will report an error and return as 0. 
It should be noted that at this point all bracket content is ALREADY EVALUATED. eg. sin(3*4) would reach here as sin12.
If functions take multiple input use ',' to divide inputs.
@style: Comment thouroughly!

Thought: Use terneary ? : operator?
*/
	double special(fstring input) { /** @code Add to this function! Be careful where you put stuff! */
		if       (input.contains_only( (char*) (s_base + ".-").str().c_str())){//normal number 
			if (s_base == "0123456789"){return input.dbl(); }
			return input.base(s_base);
		}                                                                                      /**CORE MATH*/
		else if (input.get_first(3) == "sin"    ){ return   sin(special(input.get_from(3))/180*M_PI);} /*sin*/
		else if (input.get_first(3) == "cos"    ){ return   cos(special(input.get_from(3))/180*M_PI);} /*cos*/
		else if (input.get_first(3) == "tan"    ){ return   tan(special(input.get_from(3))/180*M_PI);} /*log*/
		else if (input.get_first(4) == "asin"   ){ return  asin(special(input.get_from(4)))*180/M_PI;} /*asin*/
		else if (input.get_first(4) == "acos"   ){ return  acos(special(input.get_from(4)))*180/M_PI;} /*acos*/
		else if (input.get_first(4) == "atan"   ){ return  asin(special(input.get_from(4)))*180/M_PI;} /*asin*/
		else if (input.get_first(3) == "log"    ){ return log10(special(input.get_from(3)));} /*log_10*/
		else if (input.get_first(4) == "sqrt"   ){ return  sqrt(special(input.get_from(4)));} /*square root*/
		else if (input[0] == 'd'                ){ return rand()%(input.get_from(1)).itg()+1;} /*dice*/
		else if (input[input.length()-1] == '!' ){return factorial((int) special(input.get_first(input.length()-1)));} /*factorial*/
		                                                                                      /**Constants*/
		else if (input == "p" || input == "pi"  ){ return 3.14159;                          } /*Pi*/
		else if (input[0] =='m'||input[0] =='M'){ return get_element_mass(input.get_from(1).str());} /*get mass*/
		else if (input == "AC"                  ){ return 6.02e+23;                         }   /*Avagadro's Constant*/
		else if (input == "LperM"               ){ return 22.4;                             }   /*Litres per Mole at STP*/
		else if (input == "R"                   ){ return 8.31;                             }   /*ideal coeficient*/
		else if (input == "StP" || input =="sTp"){ return 273;                              }   /*standard temperature*/
		else if (input == "STp" || input =="stP"){ return 101.3;                            }   /*standard preasure*/
		else if (input == "g"                   ){ return 9.8;                              } /*g*/
		else if (input == "G"                   ){ return 6.673e-11;                        } /*G*/
		else if (input == "c"                   ){ return 2.998e+8;                         } /*c*/
		else if (input == "e_Q"                 ){ return 1.602e-19;                        } /*e_Q*/
		else if (input == "e"                   ){ return 2.718;                            } /*e*/
		                                                                                      /**variables*/
		else if (input == "var"                 ){ return this->var;                        } /*var*/
		else if (input[0] == 'v'                ){ return this->v[input.get_from(1).itg()]; } /*vx*/
		else if (input == "s" || input == "S"   ){ return this->s;                          } /*Stream..*/
		                                                                                      /**Unit Conversion*/
		else if (input.get_first(7) == "KMHtoMS"){ return special(input.get_from(7))*3.6;   } /*KM/H to m/s*/
		else if (input.get_first(7) == "MStoKMH"){ return special(input.get_from(7))/3.6;   } /*m/s to KM/H*/
		else if (input.get_first(4) == "CtoK"   ){ return special(input.get_from(4)) + 273; } /*C to K*/
		else if (input.get_first(4) == "KtoC"   ){ return special(input.get_from(4)) - 273; } /*K to */
		                                                                                      /**Bolean*/
		else if (input[1] == '&' && input[2] == '&'){                                         //AND
			if  (input[0] == '0' || input[3] == '0')  { return 0; } 
			else/*        both are 1                */{ return 1; }
		}else if (input[1] == '|' && input[2] == '|'){                                         //OR 
			if   (input[0] == '1' || input[3] == '1') { return 1; } 
			else/*        both are 0                */{ return 0; }
		}else if (input[0] == '!'                ){                                            //NOT
			if   (input[1] == '0'                ){ return 1; }
			else/*input[1] == '1'               */{return  0; }
		                                                                                       /**OTHER*/
		}else if (input[0] == 'F'){                                                            //Function
			double dbl_input = special(input.get_from(1));
			fstring temporary = this->F.get_till('x');
			while (!this->F.atend()){
				this->F.skip();
				temporary += dbl_input;
				temporary += this->F.get_till('x');
			}
			this->F.position = 0;
			return arithmatic(temporary);
		}else if (input.contains('@')         ) {                                              //special base specifier thing.
			fstring temporary;
			temporary = input.get_till('@');
			input.skip(); //go past @
			return temporary.base(input.get_rest());
		}else                                   {                                             //It doesn't match anything, quick, call an error!
			this->error = true;
			return 0;
		}
	}

//	map<fstring, double> user_variables;
//	map<fstring, fstring> user_functions;

	public:

	bool error;
	double var;
	double v[10];
	double s;
	fstring s_base;
	fstring F;

	solver() {
		s_base = "0123456789";
	}

	string solve(string input) {
		this->error = false;
		fstring temp;
		temp = input;
		return clean_whitespace(temp).str();
	}
};
